from __future__ import annotations

from pathlib import Path
from typing import Union, List, Dict, Any

from .utils.uni_loader import UniLoader
from .utils.uni_logger import UniLogger
from .utils.uni_saver import UniSaver
from .utils.uni_traverser import UniTraverser
from .utils.uni_traverser import traverses as _onestep_traverse
from .utils.uni_resizer import UniResizer
from .utils.uni_merger import UniMerger
from .utils.uni_peeker import UniPeeker
from .utils.utils import is_s3_uri, is_url
from .utils import constants  # from unibox.constants import IMG_FILES
from .utils.constants import *




def loads(file_path: str | Path, debug_print=True) -> any:
    """
    Loads arbitrary data from the given file path, using UniLoader.loads() method.
    :param file_path: Path to the file to load.
    :param debug_print: Whether to print debug messages. (advised to turn off if loading many files)

    example:
    >>> df = unibox.loads("data.csv")
    >>> json_dict_list = unibox.loads("data.jsonl")
    >>> pil_iimage = unibox.loads("image.png")
    """
    loader = UniLoader(debug_print=debug_print)
    return loader.loads(file_path)


def saves(data: Any, file_path: Path | str) -> None:
    """
    Saves arbitrary data to the given file path, using the UniSaver.saves() method.
    :param data: The data to saves.
    :param file_path: Path to the file to saves.

    example:
    >>> json_dict = {"name": "John", "age": 30}
    >>> unibox.saves(json_dict, "data.json")
    >>> json_dict_list = [{"name": "John", "age": 30}, {"name": "Doe", "age": 40}]
    >>> unibox.saves(json_dict_list, "data.jsonl")
    >>> df = pd.DataFrame({'A': [1, 2], 'B': [3, 4]})
    >>> unibox.saves(df, "data.parquet")
    >>> pil_image = Image.new('RGB', (60, 30), color='red')
    >>> unibox.saves(pil_image, "image.png")
    """
    saver = UniSaver()
    saver.saves(data, file_path)


def traverses(root_dir: str, include_extensions: List[str] = None,
              exclude_extensions: List[str] = None, relative_unix=False):
    """

    Args:
        root_dir: the root s3_uri to traverse
        include_extensions: list of extensions that will be included in the traversal (.txt .jpg .webp)
        exclude_extensions: list of extensions that will be excluded in the traversal (.txt .jpg .webp)
        relative_unix: whether to give a relative path or not (default False gives absolute path)

    Returns:
        list of files that were traversed
    """

    if is_s3_uri(root_dir):
        from .utils.s3_client import S3Client
        s3_client = S3Client()

        # either files or folders
        all_entries = s3_client.traverse(root_dir, include_extensions, exclude_extensions, relative_unix)
    else:
        all_entries = _onestep_traverse(root_dir, include_extensions, exclude_extensions, relative_unix)

    return all_entries


def merges(*data: Union[str, Dict, List[Any], Any]) -> Any:
    """
    Merges arbitrary data using UniMerger.merges() method.
    :param data: A variable number of data entries to merge. Each entry can be a dictionary, a list, a dataframe, or a string representing a file path.
    :param debug_print: Whether to print debug messages. (advised to turn off if merging many data entries)
    :return: Merged data
    example:
    >>> merged_dict = merges(dict1, dict2, dict3)
    >>> merged_df = merges(df1, df2, df3)
    >>> merged_data_from_files = merges("data1.csv", "data2.json", "data3.parquet")
    """
    merger = UniMerger()
    return merger.merges(*data)


def peeks(data: Any, n=3, console_print=False) -> Dict[str, Any]:
    """
    Peeks into arbitrary data using UniPeeker.peeks() method.
    :param data: The data to peek into.
    :param n: The number of entries to peek into.
    :param console_print: Whether to print the peeked information to the console.
    :return: A dictionary containing the metadata and the preview of the data.
    example:
    >>> json_dict = {"name": "John", "age": 30}
    >>> peeked_dict = peeks(json_dict)
    >>> json_dict_list = [{"name": "John", "age": 30}, {"name": "Doe", "age": 40}]
    >>> peeked_dict_list = peeks(json_dict_list)
    >>> df = pd.DataFrame({'A': [1, 2], 'B': [3, 4]})
    >>> peeked_df = peeks(df)
    """
    peeker = UniPeeker(n, console_print)
    return peeker.peeks(data)


from typing import List, Callable, Optional
from tqdm import tqdm
import os


class UniTraverser:
    """
    Universal directory traverser that handles traversal with optional file extension filtering.

    Attributes:
        root_dir: The root directory to traverse.
        include_extensions: Optional list of file extensions to include.
        exclude_extensions: Optional list of file extensions to exclude.
        pre_process: Optional function to call before traversal.
        post_process: Optional function to call after traversal.
        error_handler: Optional function to handle errors during traversal.
    """

    def __init__(self,
                 root_dir: str,
                 include_extensions: List[str] = None,
                 exclude_extensions: List[str] = None,
                 pre_process: Callable[[], None] = None,
                 post_process: Callable[[], None] = None,
                 error_handler: Callable[[Exception], None] = None):

        # Expand user and variable paths
        self.root_dir = os.path.expanduser(os.path.expandvars(root_dir))
        self.include_extensions = tuple(include_extensions) if include_extensions else None
        self.exclude_extensions = tuple(exclude_extensions) if exclude_extensions else None
        self.pre_process = pre_process
        self.post_process = post_process
        self.error_handler = error_handler
        self.traversed_files = []

    def _files_to_traverse(self):
        for root, _, files in os.walk(self.root_dir):
            for file_name in files:
                extension = os.path.splitext(file_name)[1].lower()
                if (self.include_extensions is None or extension in self.include_extensions) and \
                        (self.exclude_extensions is None or extension not in self.exclude_extensions):
                    yield os.path.join(root, file_name)

    def traverse(self, file_handler: Callable[[str], None] = None) -> None:
        """
        Traverses the directory tree and applies the given file handler to each file.

        Args:
            file_handler: A function that takes a file path and performs the desired action.
        """
        if not file_handler:
            file_handler = lambda x: x  # do nothing; only get the file paths

        if self.pre_process:
            self.pre_process()

        with tqdm(desc="Traversing files", leave=False, unit="files", total=None) as pbar:
            for file_path in self._files_to_traverse():
                try:
                    file_handler(file_path)
                except Exception as e:
                    if self.error_handler:
                        self.error_handler(e)
                    else:
                        raise e
                pbar.update(1)
                self.traversed_files.append(file_path)

        if self.post_process:
            self.post_process()

    def to_relative_unix_path(self, absolute_path: str, convert_slash: bool = True) -> str:
        """
        Converts the absolute Windows path to a relative path with respect to the root directory.
        Optionally, converts the backslashes to forward slashes for cross-platform compatibility.

        Args:
            absolute_path: The absolute Windows path to convert.
            convert_slash: Whether to convert backslashes to forward slashes.

        Returns:
            The relative Linux-like path.
        """
        # Expand user and variable paths
        absolute_path = os.path.expanduser(os.path.expandvars(absolute_path))

        relative_path = os.path.relpath(absolute_path, self.root_dir)
        if convert_slash:
            relative_path = relative_path.replace("\\", "/")
        return relative_path

    def get_traversed_files(self, relative_unix=False):
        if relative_unix:
            return [self.to_relative_unix_path(file_path) for file_path in self.traversed_files]
        else:
            return self.traversed_files


def traverses(root_dir: str, include_extensions: List[str] = None,
              exclude_extensions: List[str] = None, relative_unix=False):
    """

    Args:
        root_dir: the root s3_uri to traverse
        include_extensions: list of extensions that will be included in the traversal (.txt .jpg .webp)
        exclude_extensions: list of extensions that will be excluded in the traversal (.txt .jpg .webp)
        relative_unix: whether to give a relative path or not (default False gives absolute path)

    Returns:
        list of files that were traversed
    """
    traverser = UniTraverser(root_dir, include_extensions, exclude_extensions)
    traverser.traverse()
    files = traverser.get_traversed_files(relative_unix=relative_unix)
    return files


# example code
def pre_process():
    print("Starting traversal...")


def post_process():
    print("Finished traversal.")


def error_handler(e: Exception):
    print(f"An error occurred: {e}")


def print_txt_file(file_path: str):
    # demo file handler; does nothing but traversed file is in traverser.traversed_files
    pass


if __name__ == '__main__':
    root_dir = r"E:\Datasets\bp"
    include_extensions = [".txt"]
    traverser = UniTraverser(root_dir, include_extensions=include_extensions,
                             pre_process=pre_process, post_process=post_process,
                             error_handler=error_handler)
    traverser.traverse(print_txt_file)

    print(f"traversed {len(traverser.traversed_files)} files")
    print("Done")



import os
from urllib.parse import urlparse
def to_relaive_unix_path(absolute_path:str, root_dir:str, convert_slash=True):
    relative_path = os.path.relpath(absolute_path, root_dir)
    if convert_slash:
        relative_path = relative_path.replace("\\", "/")
    return relative_path


def is_s3_uri(uri: str) -> bool:
    """Check if the URI is an S3 URI."""
    parsed = urlparse(uri)
    return parsed.scheme == 's3'

def is_url(path: str) -> bool:
    try:
        result = urlparse(path)
        return all([result.scheme, result.netloc])
    except:
        return False


# Adding the _load_csv and _load_parquet methods to the UniLoader class

from __future__ import annotations

import json
import timeit

import os
import mimetypes
import tempfile
import requests

import tomli
import pandas as pd

from PIL import Image
from pathlib import Path
from omegaconf import OmegaConf

from .uni_logger import UniLogger
from .s3_client import S3Client
from .utils import is_url, is_s3_uri


class UniLoader:
    """A simple utility class for loading various file types.
    Data cleaning should be done in the class that uses this loader.
    """

    def __init__(self, logger=None, debug_print=True):
        if not logger:
            self.logger = UniLogger(file_suffix="UniLoader")
        else:
            self.logger = logger

        self.debug_print = debug_print

        # Mapping of file extensions to loader functions
        self.loaders = {
            '.json': self._load_json,
            '.jsonl': self._load_jsonl,  # Added '.jsonl' to the loaders dictionary
            '.txt': self._load_txt,
            '.html': self._load_txt,
            '.csv': self._load_csv,
            '.png': self._load_image,
            '.jpg': self._load_image,
            '.webp': self._load_image,
            '.jpeg': self._load_image,
            '.toml': self._load_toml,
            '.yaml': self._load_yaml,
            '.parquet': self._load_parquet,
        }

    def _load_from_s3(self, s3_uri: str):
        """Download a file from an S3 URI and load its content."""
        s3_client = S3Client()
        with tempfile.TemporaryDirectory() as tmp_dir:
            local_path = Path(s3_client.download(s3_uri, tmp_dir))
            return self.loads(local_path)

    def _load_from_url(self, url: str):
        """Download a file from a URL and load its content."""
        response = requests.get(url)
        response.raise_for_status()  # Ensure the request was successful

        _, url_suffix = os.path.splitext(url)
        if url_suffix == '':
            content_type = response.headers.get('content-type')
            mime_suffix = mimetypes.guess_extension(content_type)
            suffix = mime_suffix if mime_suffix else ''
        else:
            suffix = url_suffix

        with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as tmp_file:
            tmp_file.write(response.content)
            tmp_file.flush()
            tmp_file_path = tmp_file.name
            tmp_file.close()
            return self.loads(tmp_file_path)

    def loads(self, file_path: Path | str, encoding="utf-8"):
        """Load data from the given file path.

        The type of data returned depends on the file extension.
        """
        start_time = timeit.default_timer()

        # check if is s3 uri or url (downloads the file)
        if isinstance(file_path, str):
            if is_url(file_path):
                return self._load_from_url(file_path)
            elif is_s3_uri(file_path):
                return self._load_from_s3(file_path)

        # is a local file
        if isinstance(file_path, str):
            file_path = Path(file_path)

        file_type = file_path.suffix.lower()

        if not file_path.exists():
            self.logger.warning(f'{file_type} DOES NOT EXIST at "{file_path}"')
            return None

        if file_type not in self.loaders:
            self.logger.error(f'Unsupported file type "{file_type}"')
            return None

        try:
            # read the file using loader function
            result = self.loaders[file_type](file_path, encoding)
            if self.debug_print:
                self.logger.info(f'{file_type} LOADED from "{file_path}" in {timeit.default_timer() - start_time:.2f}s')
            return result
        except Exception as e:
            self.logger.error(f'{file_type} LOAD ERROR at "{file_path}": {e}')
            return None

    def _load_json(self, file_path: Path, encoding):
        with open(file_path, "r", encoding=encoding) as f:
            return json.load(f)

    def _load_jsonl(self, file_path: Path, encoding):
        """Load data from a .jsonl file and return it as a list of dictionaries."""
        with open(file_path, "r", encoding=encoding) as f:
            return [json.loads(line) for line in f]

    def _load_txt(self, file_path: Path, encoding):
        with open(file_path, "r", encoding=encoding) as f:
            raw_lines = f.readlines()
            return [line.strip() for line in raw_lines]

    def _load_csv(self, file_path: Path, encoding):
        return pd.read_csv(file_path, delimiter=',')

    def _load_image(self, file_path: Path, encoding):
        return Image.open(file_path)

    def _load_toml(self, file_path: Path, encoding):
        with open(file_path, "r", encoding=encoding) as f:
            return tomli.load(f)

    def _load_yaml(self, file_path: Path, encoding):
        with open(file_path, "r", encoding=encoding) as f:
            return OmegaConf.load(f)

    def _load_parquet(self, file_path: Path, encoding):
        return pd.read_parquet(file_path)


if __name__ == "__main__":
    # Usage example
    logger = UniLogger("logs", file_suffix="data_loader")
    data_loader = UniLoader(logger)
    json_data = data_loader.loads("example.json")  # string
    txt_data = data_loader.loads(Path("example.txt"))  # path
    csv_data = data_loader.loads(Path("example.csv"))
    image_data = data_loader.loads(Path("example.png"))
    toml_data = data_loader.loads(Path("example.toml"))
    yaml_data = data_loader.loads(Path("example.yaml"))
    parquet_data = data_loader.loads(Path("example.parquet"))



import os
import math
import shutil
from PIL import Image
from pathlib import Path
from tqdm.auto import tqdm

import unibox
from unibox import UniLoader, UniLogger

from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import List, Tuple


class UniResizer:
    """
    Resizes images in a directory to a specified size.

    Resolution:
        - by min_dim: resize the shorter side to a specified size
        - by max_dim: resize the longer side to a specified size
        - by target_pixels: resize the shorter side to a specified number of pixels

    Methods:
        - by directory: resize all images in a directory
        - by list: resize a list of images given

    Usage:
    >>> resizer = UniResizer(root_dir, dst_dir, min_dim=min_dim, max_dim=max_dim, target_pixels=target_pixels,
    >>>                   keep_hierarchy=False, exist_ok=True)
    >>> resizer.execute_resize_jobs(resizer.get_resize_jobs())
    """

    TRUNCATE_MULTIPLIER = 32  # the shorter side will be a multiple of 32 after scaling
    WEBP_QUALITY = 98  # quality of the output image (max:100)
    EXTENSION = ".webp"  # file extension (.webp)

    def __init__(self, root_dir: str, dst_dir: str,
                 min_dim: int = None, max_dim: int = None, target_pixels: int = None,
                 keep_hierarchy: bool = True, exist_ok: bool = True, logger: UniLogger = None):
        """
        Initialize an instance of UniResizer.

        :param root_dir: root directory containing the images to be resized
        :param dst_dir: destination directory to saves the resized images
        :param min_dim: minimum dimension of the shorter side
        :param max_dim: maximum dimension of the longer side (higher priority than min_dim)
        :param target_pixels: target number of pixels (higher priority than max_dim), will be truncated
        :param keep_hierarchy: if True, keep the original directory hierarchy; otherwise, flatten the directory
        """
        self.root_dir = root_dir
        self.dst_dir = dst_dir

        self.min_dim = min_dim
        self.max_dim = max_dim
        self.target_pixels = target_pixels

        self.keep_hierarchy = keep_hierarchy
        self.exist_ok = exist_ok

        self.logger = logger if logger is not None else UniLogger()

    @staticmethod
    def _get_new_dimensions(width: int, height: int, target_side: int, resize_by_longer_side: bool = False) -> tuple:
        """
        Calculates new dimensions based on the target,
        for either the shorter or longer side while maintaining aspect ratio.

        :param width: original width of the image
        :param height: original height of the image
        :param target_side: target dimension for the side specified
        :param resize_by_longer_side: if True, resize by longer side; otherwise, by shorter side
        :return: tuple containing new dimensions (new_width, new_height)
        """
        if target_side == -1:
            return int(width), int(height)

        if any(x <= 0 or type(x) != int for x in [width, height, target_side]):
            raise ValueError("width, height and target_side must be positive integers")

        # Determine shorter and longer side
        shorter_side, longer_side = min(width, height), max(width, height)

        if resize_by_longer_side:
            target_longer_side = target_side
            new_longer_side = target_longer_side
            new_shorter_side = round(new_longer_side * (shorter_side / longer_side))
        else:
            target_shorter_side = target_side
            new_shorter_side = target_shorter_side
            new_longer_side = round(new_shorter_side * (longer_side / shorter_side))

        # Assign new dimensions based on original orientation (portrait or landscape)
        if width > height:
            return new_longer_side, new_shorter_side
        else:
            return new_shorter_side, new_longer_side

    def _get_dst_path(self, og_rel_image_path: str) -> str:
        """
        :param og_rel_image_path: original relative image path
        """
        assert self.EXTENSION.startswith(".") and self.EXTENSION != ".", "extension must start with a dot"

        og_rel_path = Path(og_rel_image_path)
        rel_resized_path = og_rel_path.with_name(f"{og_rel_path.stem}_resized{self.EXTENSION}")

        if self.keep_hierarchy:
            dst_file_path = Path(self.dst_dir) / rel_resized_path

        else:
            dst_file_path = Path(self.dst_dir) / rel_resized_path.name

        return str(dst_file_path)

    def _create_dst_dir(self, dst_file_path: str) -> None:
        Path(dst_file_path).parent.mkdir(parents=True, exist_ok=True)

    def resize_single_image(self, image: Image.Image) -> Image.Image:
        width, height = image.size
        new_width, new_height = width, height

        # Define the need_resize variable
        need_resize = (self.min_dim is not None and min(width, height) < self.min_dim) or \
                      (self.max_dim is not None and max(width, height) > self.max_dim)

        # Priority 1: Resize based on target_pixels
        if self.target_pixels is not None:
            if width * height > self.target_pixels:  # Add this check
                scale_factor = math.sqrt(self.target_pixels / (width * height))
                target_shorter_side = round(
                    (min(width, height) * scale_factor) // self.TRUNCATE_MULTIPLIER) * self.TRUNCATE_MULTIPLIER
                new_width, new_height = self._get_new_dimensions(width, height, target_shorter_side)

        # Priority 2: Resize based on max_dim
        elif self.max_dim is not None and need_resize:
            new_width, new_height = self._get_new_dimensions(width, height, self.max_dim, resize_by_longer_side=True)

        # Priority 3: Resize based on min_dim
        elif self.min_dim is not None and need_resize:
            if min(width, height) > self.min_dim:  # Add this check
                new_width, new_height = self._get_new_dimensions(width, height, self.min_dim)

        # If resizing is not needed, return the original image
        elif not need_resize:
            return image

        # Perform the resize operation
        return image.resize((new_width, new_height), Image.LANCZOS)

    def _resize_single_image_task(self, og_rel_image_path: str) -> None:
        """
        Resize a single image and saves the result.

        :param og_rel_image_path: original image path relative to self.root_dir
        """
        loader = UniLoader(debug_print=False)
        image = loader.loads(os.path.join(self.root_dir, og_rel_image_path))
        image = self.resize_single_image(image)

        # Handle saves path
        dst_file_path = self._get_dst_path(og_rel_image_path)

        # Create destination directory
        if self.keep_hierarchy:
            self._create_dst_dir(dst_file_path)

        # Save
        try:
            with open(dst_file_path, "wb") as f:
                image.save(f, "webp", quality=self.WEBP_QUALITY)
        except OSError:
            self.logger.error(f"Error saving image {dst_file_path}. Skipping...")

    @staticmethod
    def _execute_resize_tasks(tasks: List[Tuple]):
        """
        Execute tasks using ProcessPoolExecutor
        """
        with ProcessPoolExecutor() as executor:
            futures = [executor.submit(task, *args) for task, *args in tasks]
            list(tqdm(as_completed(futures), total=len(tasks), desc="Resizing images"))

    def get_resize_jobs(self) -> List[str]:
        """
        Get the list of image files that need to be resized.

        :return: List of image paths relative to self.root_dir to be resized
        """
        self.logger.info("Getting image paths...")
        todo_image_files = unibox.traverses(self.root_dir, include_extensions=unibox.constants.IMG_FILES,
                                            relative_unix=True)

        if self.exist_ok:
            self.logger.info("Checking existing files...")
            existing_files = unibox.traverses(self.dst_dir, include_extensions=unibox.constants.IMG_FILES,
                                              relative_unix=False)

            expected_file_set = {self._get_dst_path(rel_path) for rel_path in todo_image_files}
            existing_file_set = set(existing_files)

            # Find the difference between the two sets
            needed_files_set = expected_file_set - existing_file_set

            # Map back to original relative paths
            reverse_mapping = {self._get_dst_path(rel_path): rel_path for rel_path in todo_image_files}
            todo_image_files = [reverse_mapping[dst_path] for dst_path in needed_files_set]

        return todo_image_files

    def execute_resize_jobs(self, image_files: List[str]) -> None:
        """
        Execute resizing tasks for the given list of image files.

        :param image_files: List of relative image paths to be resized
            e.g. "root_dir/some_dir/image.jpg" -> entry should be "some_dir/image_resized.jpg"
        """
        tasks = [(self._resize_single_image_task, og_rel_image_path) for og_rel_image_path in image_files]

        self.logger.info(f"Resizing {len(tasks)} images...")
        self._execute_resize_tasks(tasks)


if __name__ == '__main__':
    root_dir = r"E:\_benchmark\1k"
    dst_dir = r"E:\_benchmark\1k_resized"

    min_dim = 512  # lower priority
    max_dim = int(1024 * 3)  # higher priority
    target_pixels = int(1024 * 1024 * 1.25)  # highest priority

    resizer = UniResizer(root_dir, dst_dir, min_dim=min_dim, max_dim=max_dim, target_pixels=target_pixels,
                         keep_hierarchy=False, exist_ok=True)

    images_to_resize = resizer.get_resize_jobs()
    resizer.execute_resize_jobs(images_to_resize)



import json
import pandas as pd
from PIL import Image
from pathlib import Path
from typing import Union

from .uni_logger import UniLogger


class UniSaver:
    """A simple utility class for saving various data types to appropriate file formats.
    """

    def __init__(self, logger=None, debug_print=True):
        if not logger:
            self.logger = UniLogger(file_suffix="UniSaver")
        else:
            self.logger = logger

        self.debug_print = debug_print

    def saves(self, data, file_path: Union[Path, str]):
        """Save data to the given file path.

        The type of file saved depends on the data type.
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)

        data_type = type(data).__name__

        # Determine the expected file extension
        extension_mapping = {
            'dict': '.json',
            'list': self._list_extension(data),
            'DataFrame': '.parquet',
            'Image': '.png',
            'str': '.txt',  # Added case for strings
        }

        if data_type not in extension_mapping:
            self.logger.error(f'Unsupported data type "{data_type}"')
            return

        expected_extension = extension_mapping[data_type]

        # Verify or append the correct extension
        if file_path.suffix.lower() != expected_extension:
            if file_path.suffix:
                self.logger.error(
                    f'Invalid file extension for {data_type}. Expected "{expected_extension}" but got "{file_path.suffix}"')
                self.logger.warning("file extension will be appended to the file name")

            file_path = Path(
                str(file_path) + expected_extension)  # something.txt -> something.txt.json, so no overwrite
            self.logger.warning(f"file without extension, saving to actual path: {file_path}")

        self.handle_save(data, data_type, file_path, expected_extension)

    def _save_data(self, data, data_type, file_path, expected_extension=None):
        if data_type == 'dict':
            self._save_json(data, file_path)
        elif data_type == 'list':
            self._save_list(data, file_path, expected_extension)
        elif data_type == 'set':
            self._save_list(list(data), file_path, expected_extension)
        elif data_type == 'DataFrame':
            self._save_parquet(data, file_path)
        elif data_type == 'Image':
            self._save_image(data, file_path)
        elif data_type == 'str':  # New case for saving strings
            self._save_txt([data], file_path)  # Wrapping data in a list

        if self.debug_print:
            self.logger.info(f'{data_type} saved successfully to "{file_path}"')

    def handle_save(self, data, data_type, file_path, expected_exteison=None):
        try:
            self._save_data(data, data_type, file_path, expected_exteison)
        except PermissionError:
            alternative_file_path = file_path.replace('.parquet', '_alternative.parquet')
            self.logger.warning(
                f'Permission denied for "{file_path}". Trying to saves to "{alternative_file_path}" instead.')
            self._save_data(data, data_type, alternative_file_path, expected_exteison)
        except Exception as e:
            self.logger.error(f'{data_type} saves ERROR at "{file_path}": {e}')

    def _list_extension(self, data: list):
        if all(isinstance(item, dict) for item in data):
            return '.jsonl'
        elif all(isinstance(item, str) for item in data):
            return '.txt'
        else:
            raise TypeError("List content must be either all dictionaries or all strings.")

    def _save_json(self, data: dict, file_path: Path):
        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=2)

    def _save_list(self, data: list, file_path: Path, extension: str):
        if extension == '.jsonl':
            self._save_jsonl(data, file_path)
        elif extension == '.txt':
            self._save_txt(data, file_path)
        else:
            raise ValueError("Invalid extension for list data.")

    def _save_jsonl(self, data: list, file_path: Path):
        with open(file_path, "w", encoding="utf-8") as f:
            for item in data:
                json.dump(item, f, ensure_ascii=False)
                f.write('\n')

    def _save_txt(self, data: list, file_path: Path):
        with open(file_path, "w", encoding="utf-8") as f:
            for item in data:
                f.write(f'{item}\n')

    def _save_parquet(self, data: pd.DataFrame, file_path: Path):
        data.to_parquet(file_path, index=False)

    def _save_image(self, data: Image.Image, file_path: Path):
        data.save(file_path)


if __name__ == "__main__":
    # Usage example
    logger = UniLogger("logs", file_suffix="saver")
    saver = UniSaver(logger)
    json_data = {'key': 'value'}
    saver.saves(json_data, "example")  # Automatically adds .json
    jsonl_data = [{'key1': 'value1'}, {'key2': 'value2'}]
    saver.saves(jsonl_data, "example.jsonl")
    txt_data = ['line1', 'line2']
    saver.saves(txt_data, "example.txt")
    image_data = Image.new('RGB', (60, 30), color='red')
    saver.saves(image_data, "example.png")
    df_data = pd.DataFrame({'A': [1, 2], 'B': [3, 4]})
    saver.saves(df_data, "example.parquet")
